/*************************************************************************
* apl - abstract presentation layer                                      *
* ---------------------------------------------------------------------- *
* Copyright (C) 2000 - 2001 twisd AG, <http://www.twisd.de>              *
* See AUTHORS file for a list of contributing developers                 *
*                                                                        *
* This library is free software; you can redistribute it and/or          *
* modify it under the terms of the GNU Lesser General Public             *
* License as published by the Free Software Foundation; either           *
* version 2 of the License, or (at your option) any later version.       *
*                                                                        *
* This library is distributed in the hope that it will be useful,        *
* but WITHOUT ANY WARRANTY; without even the implied warranty of         *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      *
* Library General Public License for more details.                       *
*                                                                        *
* You should have received a copy of the GNU Lesser General Public       *
* License along with this progam; if not, write to the                   *
* Free Software Foundation, Inc., 59 Temple Place - Suite 330,           *
* Boston, MA 02111-1307, USA.                                            *
*************************************************************************/

/* $Id: apldataobject.cpp,v 1.48 2002/09/24 15:04:21 k Exp $ */


#include "apldataobject.h"
using namespace std;

CAPLDataObject::CAPLDataObject(CAPLObject* Parent) : CAPLObject(Parent) {
  iOrder = -1;
  DataLookup = NULL;
  DataControl = NULL;
  bFilter = false;
  subnode = NULL;
  setDataState(DATACONTROL_FETCH, STATE_ENABLED);
  setDataState(DATACONTROL_UPDATE, STATE_ENABLED);
}

void CAPLDataObject::setDataControl(CAPLDataControlObject *Parent) {
  Parent->addDataObject(this);
  DataControl = Parent;
}

void CAPLDataObject::setFilter(const int type, const string &filter) {
  this->filter[type] = filter;
  bFilter = true;
}

bool CAPLDataObject::getFilter(void) {
  if (bFilter)
    return true;
  else
    return false;
}

string CAPLDataObject::getFilter(const int type) {
  return filter[type];
}

void CAPLDataObject::setOrder(const int iNewOrder) {
  iOrder = iNewOrder;
}

int CAPLDataObject::getOrder(void) {
  return iOrder;
}

void CAPLDataObject::addLookup(CAPLDataLookup *DataLookup) {
  this->DataLookup = DataLookup;
}

void CAPLDataObject::setValidator(int type, apl_callback func, apl_userdat_dtor dtor, void *userdat, const string &errorMessage) {
  this->errorMessage = errorMessage;
  validator[type].setCallback(func, dtor, userdat);
}

void CAPLDataObject::executeValidator(int type, const string *key, dataset *hash, string &Error) {
  apl_dataobject_validator data;
  string sTmp = getName();

  data.type = type;
  data.key = key;
  data.field = &sTmp;
  data.hash = hash;
  data.subnode = subnode;

  data.errorMessage = NULL;

  if (validator[type].execute(&data)) {
    if (data.errorMessage) {
      Error = *data.errorMessage;
      delete data.errorMessage;
    } else {
      Error = errorMessage;
    }
  }
}

void CAPLDataObject::setConvertor(int type, apl_callback func, apl_userdat_dtor dtor, void *userdat) {
  convertor[type].setCallback(func, dtor, userdat);
}

void CAPLDataObject::executeConvertor(int type, const string *key, string *value, dataset *hash) {
  APL_TRACEIN("CAPLDataObject::executeConvertor");

  if (!DataControl->getError().length() || DataControl->getError() != *key) {
    apl_dataobject_convertor data;
    string sTmp = getName();

    data.key = key;
    data.field = &sTmp;
    data.value = value;
    data.hash = hash;
    data.keep = false;
    convertor[type].execute(&data);

    if (data.keep) setDataState(DATACONTROL_UPDATE, STATE_DISABLED);
  }

  APL_TRACEOUT("CAPLDataObject::executeConvertor");
}

void CAPLDataObject::setCallback(int type, apl_callback func, apl_userdat_dtor dtor, void *dat) {
  if (type < 0 || type >= EVENT_NUMBER) throw runtime_error("Event type specifier exceeds limit");

  event[type].setCallback(func, dtor, dat);
}

int CAPLDataObject::executeEvent(int type, const string *key, dataset *hash) {
  APL_TRACEIN("CAPLDataObject::executeEvent");
  apl_dataobject_event evdat;
  int iReturn;
  string sTmp = getName();

  evdat.key = key;
  evdat.hash = hash;
  evdat.field = &sTmp;
  evdat.subnode = subnode;
  if (hash)
    evdat.value = &(*hash)[getName()];
  else
    evdat.value = NULL;

  iReturn = event[type].execute(&evdat);

  APL_TRACEOUT("CAPLDataObject::executeEvent");
  return iReturn;
}

void CAPLDataObject::setDataState(const int datacontrol, const int state) {
  if (datacontrol == DATACONTROL_UPDATE) {
    setDataState(DATACONTROL_ADD, state);
    setDataState(DATACONTROL_MODIFY, state);
    setDataState(DATACONTROL_REMOVE, state);
  }
  if (state & 1) {
    this->datastate[datacontrol] &= (~state);
  } else {
    this->datastate[datacontrol] |= state;
  }
}

int CAPLDataObject::getDataState(const int datacontrol) {
  return datastate[datacontrol];
}

string CAPLDataObject::getValue(string primaryKey) {
  string tmpValue = CAPLObject::getValue();

  if (DataControl->getListSize() && primaryKey != "") {
    tmpValue = DataControl->DataList[primaryKey][getName()];
  }

  return tmpValue;
}

void CAPLDataObject::check(const string &object) {
  if (DataControl == NULL) throw runtime_error("DataControl was not defined for object of type " + object);
  if (getName() == "") throw runtime_error("Name was bot defined for object of type " + object);
}



CAPLDataLookup::CAPLDataLookup(CAPLDataObject *Parent) {
  if (Parent)
    Parent->addLookup(this);
  bFetched = false;
}

string CAPLDataLookup::getData(const string &sPrimaryValue) {
  check("CAPLDataLookup");

  if (!bFetched) {
    DataControl->fetch();
    bFetched = true;
  }

  return DataControl->DataList[sPrimaryValue][getName()];
}

keylist CAPLDataLookup::getList(void) {
  if (!bFetched) {
    DataControl->fetch();
    bFetched = true;
  }
  return DataControl->getList();
}

void CAPLDataLookup::addData(const string &key, const string &value) {
  if (value.length())
    Data[key] = value;
  else
    Data[key] = key;
}
