/*************************************************************************
* apl - abstract presentation layer                                      *
* ---------------------------------------------------------------------- *
* Copyright (C) 2000 - 2001 twisd AG, <http://www.twisd.de>              *
* See AUTHORS file for a list of contributing developers                 *
*                                                                        *
* This library is free software; you can redistribute it and/or          *
* modify it under the terms of the GNU Lesser General Public             *
* License as published by the Free Software Foundation; either           *
* version 2 of the License, or (at your option) any later version.       *
*                                                                        *
* This library is distributed in the hope that it will be useful,        *
* but WITHOUT ANY WARRANTY; without even the implied warranty of         *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      *
* Library General Public License for more details.                       *
*                                                                        *
* You should have received a copy of the GNU Lesser General Public       *
* License along with this progam; if not, write to the                   *
* Free Software Foundation, Inc., 59 Temple Place - Suite 330,           *
* Boston, MA 02111-1307, USA.                                            *
*************************************************************************/

/* $Id: php_apl_calls.cpp,v 1.49 2002/09/24 15:04:22 k Exp $ */


#include <stdexcept>
#include <string>
#include "php_apl.h"
#include "php_apl_calls.h"
#include "aplformcontrolobject.h"
#include "apldataobject.h"

extern "C" {
# include "php.h"
}

using namespace std;

void aplPrependDocumentRoot(const char *s, size_t len, string *result) {
  zval **var;
  zval **data;
  ELS_FETCH();

  if (zend_hash_find(&EG(symbol_table), "HTTP_SERVER_VARS", sizeof("HTTP_SERVER_VARS"), (void **) &data) == SUCCESS) {
    if (zend_hash_find(Z_ARRVAL_PP(data), "SCRIPT_FILENAME", sizeof("DOCUMENT_ROOT"), (void **) &var) == SUCCESS) {
      result->assign(Z_STRVAL_PP(var), Z_STRLEN_PP(var));
      result->erase(result->find_last_of("/") + 1);
      result->append(s, len);
      return;
    }
  }
  result->assign(s, len);
}

// Helper functions for callback

void aplCreateArguments(int n, zval **argv) {
  for (int i = 0; i < n; i++) {
    MAKE_STD_ZVAL(argv[i]);
  }
}

void aplDestroyArguments(int n, zval **argv) {
  for (int i = 0; i < n; i++) {
    zval_ptr_dtor(&argv[i]);
  }
}

void aplCallbackDestructor(void *dummy) {
  apl_php_callback_t *dat = (apl_php_callback_t *) dummy;

  zval_ptr_dtor(&dat->func);
  efree(dat);
}

static void aplCreateZvalFromHash(apl_action_hash *hash, string *value, zval *array) {
  zval *property;

  if (!hash->size()) {
    if (value) {
      ZVAL_STRINGL(array, (char *) value->c_str(), value->length(), 1);
    } else {
      ZVAL_BOOL(array, 0);
    }
    return ;
  }

  array_init(array);

  for (apl_action_hash::iterator i = hash->begin(); i != hash->end(); i++) {
    MAKE_STD_ZVAL(property);
    aplCreateZvalFromHash(&i->second.data, &i->second.value, property);
    zend_hash_update(array->value.ht, (char *) i->first.c_str(),
                     i->first.length() + 1, (void *) &property, sizeof(zval *), NULL);
  }
}

int aplCallHandler(zval *func, int argc, zval **argv, std::string **retString) {
  zval * retval;
  int ret = 0;
  char *name;
  ELS_FETCH();

  MAKE_STD_ZVAL(retval);

  if (!zend_is_callable(func, 0, &name)) {
    zend_error(E_ERROR, "User supplied callback %s not found", name);
  }

  if (call_user_function(EG(function_table), NULL, func, retval, argc, argv) == FAILURE) {
    zend_error(E_ERROR, "Unable to call %s", name);
  }

  if (name) efree(name);

  if (retString && Z_TYPE_P(retval) == IS_STRING) {
    *retString = new string;
    (*retString)->assign(Z_STRVAL_P(retval), Z_STRLEN_P(retval));
    ret = 1;
  } else {
    convert_to_long(retval);
    ret = Z_LVAL_P(retval);
  }
  zval_ptr_dtor(&retval);

  return ret;
}

string * aplCallHandlerString(zval *func, int argc, zval **argv) {
  zval *retval;
  string * ret;
  char *name;
  ELS_FETCH();

  MAKE_STD_ZVAL(retval);

  if (!zend_is_callable(func, 0, &name)) {
    zend_error(E_ERROR, "User supplied callback %s not found", name);
  }

  if (call_user_function(EG(function_table), NULL, func, retval, argc, argv) == FAILURE) {
    zend_error(E_ERROR, "Unable to call %s", name);
  }

  if (name) efree(name);

  if (Z_TYPE_P(retval) == IS_NULL) {
    ret = NULL;
  } else {
    ret = new string();
    convert_to_string(retval);
    ret->assign(Z_STRVAL_P(retval), Z_STRLEN_P(retval));
  }

  zval_ptr_dtor(&retval);

  return ret;
}

// Call Handler Implementation

int aplCallHandlerEvent(zval *func, const string *key, dataset *hash) {
  zval *args[2];
  int ret;

  aplCreateArguments(2, args);

  if (key) {
    ZVAL_STRINGL(args[0], (char *) key->c_str(), key->length(), 1);
  } else {
    ZVAL_EMPTY_STRING(args[0]);
  }

  if (hash) {
    array_init(args[1]);

    for (dataset::iterator Iterate = hash->begin(); Iterate != hash->end(); Iterate++) {
      add_assoc_string(args[1], (char *) Iterate->first.c_str(), (char *) Iterate->second.c_str(), 1);
    }
  } else {
    ZVAL_EMPTY_STRING(args[1]);
  }

  ret = aplCallHandler(func, 2, args);
  aplDestroyArguments(2, args);
  return ret;
}

int aplCallHandlerDataObject(zval *func, const string *key, const string *field, const string *value, dataset *hash, apl_action_data *action) {
  zval * args[5];
  int ret;

  aplCreateArguments(5, args);

  if (key) {
    ZVAL_STRINGL(args[0], (char *) key->c_str(), key->length(), 1);
  } else {
    ZVAL_EMPTY_STRING(args[0]);
  }

  if (field) {
    ZVAL_STRINGL(args[1], (char *) field->c_str(), field->length(), 1);
  } else {
    ZVAL_EMPTY_STRING(args[1]);
  }

  if (value) {
    ZVAL_STRINGL(args[2], (char *) value->c_str(), value->length(), 1);
  } else {
    ZVAL_EMPTY_STRING(args[2]);
  }

  if (action) {
    aplCreateZvalFromHash(&action->data, NULL, args[3]);
  } else {
    ZVAL_EMPTY_STRING(args[3]);
  }

  if (hash) {
    array_init(args[4]);

    for (dataset::iterator Iterate = hash->begin(); Iterate != hash->end(); Iterate++) {
      add_assoc_string(args[4], (char *) Iterate->first.c_str(), (char *) Iterate->second.c_str(), 1);
    }
  } else {
    ZVAL_EMPTY_STRING(args[4]);
  }

  ret = aplCallHandler(func, 5, args);
  aplDestroyArguments(5, args);
  return ret;
}


string * aplCallHandlerConvertor(zval *func, const string *key, const string *field, const string *value, dataset *hash) {
  zval *args[4];
  string * ret;

  aplCreateArguments(4, args);

  if (key) {
    ZVAL_STRINGL(args[0], (char *) key->c_str(), key->length(), 1);
  } else {
    ZVAL_EMPTY_STRING(args[0]);
  }

  if (field) {
    ZVAL_STRINGL(args[1], (char *) field->c_str(), field->length(), 1);
  } else {
    ZVAL_EMPTY_STRING(args[1]);
  }

  if (value) {
    ZVAL_STRINGL(args[2], (char *) value->c_str(), value->length(), 1);
  } else {
    ZVAL_EMPTY_STRING(args[2]);
  }

  if (hash) {
    array_init(args[3]);

    for (dataset::iterator Iterate = hash->begin(); Iterate != hash->end(); Iterate++) {
      add_assoc_string(args[3], (char *) Iterate->first.c_str(), (char *) Iterate->second.c_str(), 1);
    }
  } else {
    ZVAL_EMPTY_STRING(args[3]);
  }

  ret = aplCallHandlerString(func, 4, args);
  aplDestroyArguments(4, args);
  return ret;
}

int aplCallHandlerValidator(zval *func, const string *key, const string *field, const string *value, dataset *hash, string **retString) {
  zval *args[4];
  int ret;

  aplCreateArguments(4, args);

  if (key) {
    ZVAL_STRINGL(args[0], (char *) key->c_str(), key->length(), 1);
  } else {
    ZVAL_EMPTY_STRING(args[0]);
  }

  if (field) {
    ZVAL_STRINGL(args[1], (char *) field->c_str(), field->length(), 1);
  } else {
    ZVAL_EMPTY_STRING(args[1]);
  }

  if (value) {
    ZVAL_STRINGL(args[2], (char *) value->c_str(), value->length(), 1);
  } else {
    ZVAL_EMPTY_STRING(args[2]);
  }

  if (hash) {
    array_init(args[3]);

    for (dataset::iterator i = hash->begin(); i != hash->end(); i++) {
      add_assoc_string(args[3], (char *) i->first.c_str(), (char *) i->second.c_str(), 1);
    }
  } else {
    ZVAL_EMPTY_STRING(args[3]);
  }

  ret = aplCallHandler(func, 4, args, retString);
  aplDestroyArguments(4, args);
  return ret;
}

int aplCallHandlerValidatorCollector(zval *func, const string *key, const string *field, apl_action_data *action, dataset *hash, string **retString) {
  zval *args[4];
  int ret;

  aplCreateArguments(4, args);

  if (key) {
    ZVAL_STRINGL(args[0], (char *) key->c_str(), key->length(), 1);
  } else {
    ZVAL_EMPTY_STRING(args[0]);
  }

  if (field) {
    ZVAL_STRINGL(args[1], (char *) field->c_str(), field->length(), 1);
  } else {
    ZVAL_EMPTY_STRING(args[1]);
  }

  if (action) {
    aplCreateZvalFromHash(&action->data, NULL, args[2]);
  } else {
    ZVAL_EMPTY_STRING(args[2]);
  }

  if (hash) {
    array_init(args[3]);

    for (dataset::iterator i = hash->begin(); i != hash->end(); i++) {
      add_assoc_string(args[3], (char *) i->first.c_str(), (char *) i->second.c_str(), 1);
    }
  } else {
    ZVAL_EMPTY_STRING(args[3]);
  }

  ret = aplCallHandler(func, 4, args, retString);
  aplDestroyArguments(4, args);
  return ret;
}


static int aplCallbackValidator(void *dummy, void *dummy2) {
  apl_php_callback_arg_t *php_dat = (apl_php_callback_arg_t *) dummy;
  apl_dataobject_validator *apl_dat = (apl_dataobject_validator *) dummy2;

  switch (apl_dat->type) {
  case VALIDATOR_STANDARD:
    return aplCallHandlerValidator(php_dat->func, apl_dat->key, apl_dat->field,
                                   &(*apl_dat->hash)[*apl_dat->field], apl_dat->hash, &apl_dat->errorMessage);
    break;
  case VALIDATOR_COLLECTOR:
    return aplCallHandlerValidatorCollector(php_dat->func, apl_dat->key, apl_dat->field,
                                            apl_dat->subnode, apl_dat->hash, &apl_dat->errorMessage);
    break;
  }

  return 0;
}


ZEND_NAMED_FUNCTION(aplSetValidator) {
  CAPLDataObject *Class;
  zval **arg[3];
  apl_php_callback_arg_t *dat;

  if (zend_get_parameters_array_ex(3, arg)) WRONG_PARAM_COUNT;
  convert_to_long_ex(arg[0]);
  convert_to_string_ex(arg[2]);

  if (Z_LVAL_PP(arg[0]) < 0 || Z_LVAL_PP(arg[0]) >= VALIDATOR_NUMBER) {
    zend_error(E_ERROR, "Validator number exceeds limit");
  }

  dat = (apl_php_callback_arg_t *) emalloc(sizeof(*dat));
  ZVAL_ADDREF(*arg[1]);
  dat->func = *arg[1];

  CATCH_ERROR_BEGIN()
  Class = (CAPLDataObject *) aplGetClass(this_ptr);
  Class->setValidator(Z_LVAL_PP(arg[0]), aplCallbackValidator, aplCallbackDestructor, dat, Z_STRVAL_PP(arg[2]));
  CATCH_ERROR_END()
}

static int aplCallbackConvertor(void *dummy, void *dummy2) {

  apl_php_callback_t *php_dat = (apl_php_callback_t *) dummy;
  apl_dataobject_convertor *apl_dat = (apl_dataobject_convertor *) dummy2;
  string * retValue = NULL;

  retValue = aplCallHandlerConvertor(php_dat->func, apl_dat->key, apl_dat->field, apl_dat->value, apl_dat->hash);

  if (retValue == NULL)
    apl_dat->keep = true;
  else
    *apl_dat->value = *retValue;

  return 0;
}

ZEND_NAMED_FUNCTION(aplSetConvertor) {
  CAPLDataObject *Class;
  zval **arg[2];
  apl_php_callback_t *dat;

  if (zend_get_parameters_array_ex(2, arg)) WRONG_PARAM_COUNT;
  convert_to_long_ex(arg[0]);

  if (Z_LVAL_PP(arg[0]) < 0 || Z_LVAL_PP(arg[0]) >= CONVERTOR_NUMBER) {
    zend_error(E_ERROR, "Convertor number exceeds limit");
  }

  ZVAL_ADDREF(*arg[1]);
  dat = (apl_php_callback_t *) emalloc(sizeof(*dat));
  dat->func = *arg[1];
  CATCH_ERROR_BEGIN()
  Class = (CAPLDataObject *) aplGetClass(this_ptr);
  Class->setConvertor(Z_LVAL_PP(arg[0]), aplCallbackConvertor, aplCallbackDestructor, dat);
  CATCH_ERROR_END()
}


static int aplDataObjectCallbackEvent(void *dummy, void *dummy2) {
  apl_php_callback_t *php_dat = (apl_php_callback_t *) dummy;
  apl_dataobject_event *apl_dat = (apl_dataobject_event *) dummy2;

  return aplCallHandlerDataObject(php_dat->func, apl_dat->key, apl_dat->field,
                                  apl_dat->value, apl_dat->hash, apl_dat->subnode);
}


ZEND_NAMED_FUNCTION(aplSetDataObjectCallback) {
  CAPLDataObject *Class;
  zval **arg[2];
  apl_php_callback_t *dat;

  if (zend_get_parameters_ex(2, &arg[0], &arg[1])) WRONG_PARAM_COUNT;
  convert_to_long_ex(arg[0]);
  if (Z_LVAL_PP(arg[0]) >= EVENT_NUMBER || Z_LVAL_PP(arg[0]) < 0) {
    php_error(E_WARNING, "Event number exceeds limit");
    return ;
  }
  ZVAL_ADDREF(*arg[1]);
  dat = (apl_php_callback_t *) emalloc(sizeof(*dat));
  dat->func = *arg[1];
  CATCH_ERROR_BEGIN()
  Class = (CAPLDataObject *) aplGetClass(this_ptr);
  Class->setCallback(Z_LVAL_PP(arg[0]), aplDataObjectCallbackEvent, aplCallbackDestructor, dat);
  CATCH_ERROR_END()
}

static int aplCallbackFailure(void *dummy, void *dummy2) {
  int ret;
  zval *args[4];
  apl_php_callback_t *php_dat = (apl_php_callback_t *) dummy;
  apl_form_failure *apl_dat = (apl_form_failure *) dummy2;

  aplCreateArguments(5, args);

  if (apl_dat->sKey) {
    ZVAL_STRINGL(args[0], (char *) apl_dat->sKey->c_str(), apl_dat->sKey->length(), 1);
  } else {
    ZVAL_EMPTY_STRING(args[0]);
  }

  if (apl_dat->sField) {
    ZVAL_STRINGL(args[1], (char *) apl_dat->sField->c_str(), apl_dat->sField->length(), 1);
  } else {
    ZVAL_EMPTY_STRING(args[1]);
  }

  if (apl_dat->sValue) {
    ZVAL_STRINGL(args[2], (char *) apl_dat->sValue->c_str(), apl_dat->sValue->length(), 1);
  } else {
    ZVAL_EMPTY_STRING(args[2]);
  }

  if (apl_dat->mHash) {
    array_init(args[3]);

    for (dataset::iterator i = apl_dat->mHash->begin(); i != apl_dat->mHash->end(); i++) {
      add_assoc_string(args[3], (char *) i->first.c_str(), (char *) i->second.c_str(), 1);
    }
  } else {
    ZVAL_EMPTY_STRING(args[3]);
  }

  ret = aplCallHandler(php_dat->func, 4, args);
  aplDestroyArguments(4, args);
  return ret;
}

ZEND_NAMED_FUNCTION(aplSetFailureCallback) {
  CAPLFormControlObject *Class;
  zval **arg[2];
  apl_php_callback_t *dat;

  if (zend_get_parameters_array_ex(2, arg)) WRONG_PARAM_COUNT;
  convert_to_long_ex(arg[0]);
  if (Z_LVAL_PP(arg[0]) < 0 || Z_LVAL_PP(arg[0]) >= ACTION_NUMBER) {
    php_error(E_WARNING, "Action number exceeds limit");
    return ;
  }
  ZVAL_ADDREF(*arg[1]);
  dat = (apl_php_callback_t *) emalloc(sizeof(*dat));
  dat->func = *arg[1];

  CATCH_ERROR_BEGIN()
  Class = (CAPLFormControlObject *) aplGetClass(this_ptr);
  Class->setFailureCallback(Z_LVAL_PP(arg[0]), aplCallbackFailure, aplCallbackDestructor, dat);
  CATCH_ERROR_END()
}

static int aplCallbackAction(void *dummy, void *dummy2) {
  apl_php_callback_t *php_dat = (apl_php_callback_t *) dummy;
  apl_formcontrol_action *apl_dat = (apl_formcontrol_action *) dummy2;

  return aplCallHandlerEvent(php_dat->func, apl_dat->sKey, apl_dat->mHash);
}

ZEND_NAMED_FUNCTION(aplSetActionCallback) {
  CAPLFormControlObject *Class;
  zval **arg[2];
  apl_php_callback_t *dat;

  if (zend_get_parameters_array_ex(2, arg)) WRONG_PARAM_COUNT;
  convert_to_long_ex(arg[0]);
  if (Z_LVAL_PP(arg[0]) < 0 || Z_LVAL_PP(arg[0]) >= ACTION_NUMBER) {
    php_error(E_WARNING, "Action number exceeds limit");
    return ;
  }
  ZVAL_ADDREF(*arg[1]);
  dat = (apl_php_callback_t *) emalloc(sizeof(*dat));
  dat->func = *arg[1];
  CATCH_ERROR_BEGIN()
  Class = (CAPLFormControlObject *) aplGetClass(this_ptr);
  Class->setCallback(Z_LVAL_PP(arg[0]), aplCallbackAction, aplCallbackDestructor, dat);
  CATCH_ERROR_END()
}

